#!/usr/bin/env python

"""Console script for Kyos."""

from __future__ import print_function
from __future__ import absolute_import

import argparse
import logging
import sys

from kyos import features
from kyos import nn
from kyos.__init__ import __version__


def parse_arguments(system_args):
    """Parse command line arguments.

    Parameters
    ----------
    system_args : list
        List of command line arguments, usually sys.argv[1:].

    Returns
    -------
    Namespace
        Command line arguments are stored as attributes of a Namespace.
    """
    def non_negative_int(value):
        try:
            ivalue = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError("Must be a number >= 0")
        if ivalue < 0:
            raise argparse.ArgumentTypeError("Must be >= 0")
        return ivalue

    def positive_int(value):
        try:
            ivalue = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError("Must be a number greater than 0")
        if ivalue <= 0:
            raise argparse.ArgumentTypeError("Must be greater than 0")
        return ivalue

    description = """Tools for haploid variant calling with Deep Neural Networks."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--version", action="version", version="%(prog)s version " + __version__)
    subparsers = parser.add_subparsers(dest="subparser_name", help=None, metavar="subcommand")
    subparsers.required = True

    formatter_class = argparse.ArgumentDefaultsHelpFormatter

    '''help_str = "Extract and tabulate informative features from a BAM file."
    description = help_str + "  Pileup positions with evidence of variation are written to the output file."
    subparser = subparsers.add_parser("extract", formatter_class=formatter_class, description=description, help=help_str)
    subparser.add_argument(dest="bam_path", type=str, metavar="BAMFILE", help="Input sorted BAM file.")
    subparser.add_argument(dest="ref_path", type=str, metavar="REFFILE", help="Input reference fasta file.")
    subparser.add_argument("--truth", dest="truth_file_path", type=str, metavar="TRUTH", help="Optional input truth file.  When specified, the last column of the output TSV file will be the truth call.")
    subparser.add_argument("--tnfract", dest="tn_fract", type=float, default=1.0, metavar="FRACTION", help="Fraction of true negative observations to retain in the output.")
    subparser.add_argument(dest="output_path", type=str, metavar="OUTFILE", help="Output tabulated feature file.")
    subparser.set_defaults(func=extract_command)

    description = "Merge multiple tabulated files and write to stdout."
    subparser = subparsers.add_parser("merge", formatter_class=formatter_class, description=description, help=description)
    subparser.add_argument(dest="input_paths", type=str, metavar="INFILE", help="Input tabulated feature files.", nargs='+')
    subparser.set_defaults(func=merge_command)'''

    help_str = "Extract and tabulate informative features from a BAM file."
    description = """Generate tabular data containing places of mutations from a reference genome. Takes a BAM
           file. Outputs the positions, number of forward and reverse ACTG, deletions, insertions and
           reference skipped, and the qualities of the ACTG SNP's."""

    subparser = subparsers.add_parser("tabulate", formatter_class=formatter_class, description=description, help=help_str)

    subparser.add_argument(dest="input_file",       type=str,    help="Input sorted bam file.")
    subparser.add_argument(dest="output_file",      type=str,    help="Output tabular data file.")
    subparser.add_argument(dest="ref_file",         type=str,    help="Input reference file.")
    subparser.add_argument("-t", "--truth", dest="truth_file", type=str, help="SNP Mutator summary file containing truth allele.", default=None)
    subparser.add_argument("-f", "--tnfract", dest="tnfract", type=float, help="Fraction of True Negatives to write to the output file.", default=1.0)
    subparser.add_argument("-s", "--rseed", dest="rseed", type=int, help="Random seed to ensure reproducible results when using --tnfract.  Set to zero for non-deterministic results.", default=1)
    subparser.add_argument("--force_truth", dest="force_truth", action='store_true', help="Extract all known variant observations even if there is no coverage. Requires a summary file.", default=False)
    subparser.set_defaults(func=tabulate_command)

    description = "Merge multiple tabulated files and write to stdout."
    subparser = subparsers.add_parser("merge", formatter_class=formatter_class, description=description, help=description)
    subparser.add_argument(dest="input_paths", type=str, metavar="INFILE", help="Input tabulated feature files.", nargs='+')
    subparser.set_defaults(func=merge_command)

    description = "Train a neural network to detect variants."
    subparser = subparsers.add_parser("train", formatter_class=formatter_class, description=description, help=description)
    subparser.add_argument(dest="train_file_path", type=str, metavar="FTR", help="Input tabulated feature file for training.")
    subparser.add_argument(dest="validate_file_path", type=str, metavar="FTR", help="Input tabulated feature file for validation during training.")
    subparser.add_argument(dest="model_file_path", type=str, metavar="MODEL", help="Output trained model.")
    subparser.add_argument("--rseed", dest="rseed", type=int, help="Random seed to ensure reproducible results.  Set to non-zero value for reproducible results.", default=0)
    subparser.set_defaults(func=train_command)

    description = "Test a neural network model when the truth is known."
    subparser = subparsers.add_parser("test", formatter_class=formatter_class, description=description, help=description)
    subparser.add_argument(dest="model_file_path", type=str, metavar="MODEL", help="Input trained model.")
    subparser.add_argument(dest="test_file_path", type=str, metavar="FTR", help="Input tabulated feature file for testing.")
    # subparser.add_argument("--vcf", dest="vcf_file_path", type=str, metavar="VCF", help="Optional output VCF file.")
    subparser.set_defaults(func=test_command)

    description = "Call variants when the truth is unknown."
    subparser = subparsers.add_parser("call", formatter_class=formatter_class, description=description, help=description)
    subparser.add_argument(dest="model_file_path", type=str, metavar="MODEL", help="Input trained model.")
    subparser.add_argument(dest="ftr_file_path", type=str, metavar="FTR", help="Input tabulated feature file.")
    subparser.add_argument(dest="vcf_file_path", type=str, metavar="VCF", help="Output VCF file.")
    subparser.set_defaults(func=call_command)

    args = parser.parse_args(system_args)
    return args


def train_command(args):
    """Train a neural network to detect variants.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    """
    nn.train(args.train_file_path, args.validate_file_path, args.model_file_path, args.rseed)


def test_command(args):
    """YYY multiple files and write to stdout.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    """
    nn.test(args.model_file_path, args.test_file_path)  # , args.vcf_file_path)


def call_command(args):
    """YYY multiple files and write to stdout.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    """
    nn.call(args.model_file_path, args.ftr_file_path, args.vcf_file_path)


def merge_command(args):
    """Merge multiple tabulated files and write to stdout.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    """
    features.merge(args.input_paths)


def tabulate_command(args):
    """YYY multiple files and write to stdout.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    """
    if args.tnfract < 1.0 and args.truth_file is None:
        print("You must specify a truth file when requesting a subset of the true negative observations.", file=sys.stderr)
        exit(1)
    features.create_tabular_data(args.input_file, args.output_file, args.ref_file, args.truth_file, args.tnfract, args.force_truth, args.rseed)


def run_command_from_args(args):
    """Run a subcommand with previously parsed arguments in an argparse namespace.

    This function is intended to be used for unit testing.

    Parameters
    ----------
    args : Namespace
        Command line arguments are stored as attributes of a Namespace.
        The args are obtained by calling parse_argument_list().

    Returns
    -------
    Returns 0 on success if it completes with no exceptions.
    """
    return args.func(args)  # this executes the function previously associated with the subparser with set_defaults


def run_from_line(line):
    """Run a command with a command line.

    This function is intended to be used for unit testing.

    Parameters
    ----------
    line : str
        Command line.

    Returns
    -------
    Returns 0 on success if it completes with no exceptions.
    """
    argv = line.split()
    args = parse_arguments(argv)
    return args.func(args)  # this executes the function previously associated with the subparser with set_defaults


def main():
    """This is the main function which is turned into an executable
    console script by the setuptools entry_points.  See setup.py.

    To run this function as a script, first install the package:
        $ python setup.py develop
        or
        $ pip install --user kyos

    Parameters
    ----------
    This function must not take any parameters

    Returns
    -------
    The return value is passed to sys.exit().
    """
    enable_log_timestamps = True
    if enable_log_timestamps:
        logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.DEBUG)
    else:
        logging.basicConfig(format="%(message)s", level=logging.DEBUG)
    args = parse_arguments(sys.argv[1:])
    return args.func(args)  # this executes the function previously associated with the subparser with set_defaults


# This snippet lets you run the cli without installing the entrypoint.
if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""This module is the neural network portion of Kyos.
"""

from __future__ import print_function
from __future__ import absolute_import

import keras
from keras.models import Sequential
from keras.layers import Dense, Activation, Dropout
from keras.callbacks import EarlyStopping
import logging
import pandas as pd
import random
import numpy as np
import tensorflow as tf
from timeit import default_timer as timer

from kyos import features
from kyos.__init__ import __version__


# Zero-based index to first feature column.
first_ftr_idx = features.feature_names.index(features.first_ftr_name)

# Zero-based index to last feature column.
last_ftr_idx = features.feature_names.index(features.last_ftr_name)

num_input_features = 1 + last_ftr_idx - first_ftr_idx

# the truth column is appended to feature_names when the truth file is provided
target_label_name = features.target_label_name

output_classes = ["A", "T", "C", "G", "A_insertion", "T_insertion", "C_insertion", "G_insertion", "_deletion"]


def relevant_data(data, first_col, last_col):
    """Extract the specified range of adjacent feature columns from a single observation row.

    Parameters
    ----------
    data : list
        List of feature and target columns.
    first_col : int
        First column index to select.
    last_col : int
        Last column index to select.

    Examples
    --------
    >>> relevant_data(list(range(0, 30)), 3, 28)
    [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0, 21.0, 22.0, 23.0, 24.0, 25.0, 26.0, 27.0, 28.0]
    """
    z = [float(val) for val in data[first_col: 1 + last_col]]
    return z


def conv_allele(allele):
    if allele.upper() == "A":
        return 0
    elif allele.upper() == "T":
        return 1
    elif allele.upper() == "C":
        return 2
    elif allele.upper() == "G":
        return 3
    elif "_insertion" in allele:
        if allele[0].upper() == "A":
            return 4
        elif allele[0].upper() == "T":
            return 5
        elif allele[0].upper() == "C":
            return 6
        elif allele[0].upper() == "G":
            return 7
    elif "_deletion" in allele:
        return 8
    else:
        raise ValueError("Unknown allele: %s" % (allele))


def conv_output(output):
    for x in range(len(output)):
        if output[x] > 0.7:
            return output_classes[x]

    return "-"


def normalize_features(df):
    """Normalize the input features so the scaled features are scaled approximately in the range 0 to 1.

    In practice, we found better results when the read count features are scaled in such a way that some
    of the values are greater than 1.

    The datafame is modified in-place.

    Parameters
    ----------
    df : Pandas dataframe
        Dataframe of features
    """
    # The scaling parameters below are hard-coded to be sure the same values will be used for training, testing,
    # and calling variants.  It might happen that new datasets will have different distribution of values.
    for ftr_name in features.read_count_feature_names:
        df[ftr_name] = df[ftr_name] / 20.0  # 60th percentile, so some values will scale higher than 1
    for ftr_name in features.map_quality_feature_names:
        df[ftr_name] = df[ftr_name] / 100.0
    for ftr_name in features.base_quality_feature_names:
        df[ftr_name] = df[ftr_name] / 100.0


def standardize_features(df):
    """Standardize the input features so the scaled features are approximately centered around 0 with variance 1.

    The datafame is modified in-place.

    Parameters
    ----------
    df : Pandas dataframe
        Dataframe of features
    """
    # The scaling parameters below are hard-coded to be sure the same values will be used for training, testing,
    # and calling variants.  It might happen that new datasets will have different distribution of values.
    for ftr_name in features.read_count_feature_names:
        df[ftr_name] = (df[ftr_name] - 6.0) / 11.0
    for ftr_name in features.map_quality_feature_names:
        df[ftr_name] = (df[ftr_name] - 50.0) / 50.0
    for ftr_name in features.base_quality_feature_names:
        df[ftr_name] = (df[ftr_name] - 50.0) / 50.0


def load_train_data(path, first_ftr_col, last_ftr_col, scaling="normalize"):
    """Load training file with features and target class.

    The features must be adjacent columns.
    The target must be the last column.
    It is okay if there are some unused columns before the first feature and after the last feature.

    Parameters
    ----------
    path : str
        Path to tabulated feature file.
    first_ftr_col : int
        Zero-based index to first feature column.
    first_ftr_col : int
        Zero-based index to last feature column.
    scaling : str, optional
        String specifying how to scale the input features.
        Possible values are: "normalize" and "standardize".  By default, features are normalized.

    Returns
    -------
    features : nparray
        Two dimensional array of features with one row per observation and one column per feature
    one_hot_labels : nparray
        One hot encoded target labels with one row per observation and one column per output class.
        All columns will be 0 except the target column will be 1.
    """
    start = timer()

    with open(path, "r") as csv_file:
        logging.debug("Reading tsv file...")
        feature_columns = features.feature_names[first_ftr_col: 1 + last_ftr_col]
        usecols = feature_columns + [target_label_name]
        dtype = {col: np.float32 for col in feature_columns}
        converters = {target_label_name: conv_allele}
        df = pd.read_csv(csv_file, sep='\t', usecols=usecols, dtype=dtype, converters=converters)

    if scaling == "normalize":
        logging.debug("Normalizing features...")
        normalize_features(df)
    elif scaling == "standardize":
        logging.debug("Standardizing features...")
        standardize_features(df)

    logging.debug("Extracting columns to np arrays...")
    data = df[feature_columns].values
    labels = df[target_label_name].values

    logging.debug("One-hot encoding target labels...")
    one_hot_labels = keras.utils.to_categorical(labels, num_classes=9)

    end = timer()
    logging.debug("%.1f seconds loading %d rows in file %s" % (end - start, len(df), path))
    return data, one_hot_labels


def load_test_data(path, first_ftr_col, last_ftr_col, scaling="normalize"):
    """Load testing file with features and target class.

    The features must be adjacent columns.
    The target must be the last column.
    It is okay if there are some unused columns before the first feature and after the last feature.

    Parameters
    ----------
    path : str
        Path to tabulated feature file.
    first_ftr_col : int
        Zero-based index to first feature column.
    first_ftr_col : int
        Zero-based index to last feature column.
    scaling : str, optional
        String specifying how to scale the input features.
        Possible values are: "normalize" and "standardize".  By default, features are normalized.

    Returns
    -------
    features : nparray
        Two dimensional array of features with one row per observation and one column per feature
    labels : nparray
        Integer encoded labels with one row per observation
    refs : nparray
        Integer encoded references with one row per observation
    """
    start = timer()

    with open(path, "r") as csv_file:
        logging.debug("Reading tsv file...")
        feature_columns = features.feature_names[first_ftr_col: 1 + last_ftr_col]
        usecols = feature_columns + [target_label_name] + ["RefBase"]
        dtype = {col: np.float32 for col in feature_columns}
        converters = {target_label_name: conv_allele, "RefBase": conv_allele}
        df = pd.read_csv(csv_file, sep='\t', usecols=usecols, dtype=dtype, converters=converters)

    if scaling == "normalize":
        logging.debug("Normalizing features...")
        normalize_features(df)
    elif scaling == "standardize":
        logging.debug("Standardizing features...")
        standardize_features(df)

    logging.debug("Extracting columns to np arrays...")
    data = df[feature_columns].values
    labels = df[target_label_name].values
    refs = df["RefBase"].values

    end = timer()
    logging.debug("%.1f seconds loading %d rows in file %s" % (end - start, len(df), path))
    return data, labels, refs


def train(train_file_path, validate_file_path, model_file_path, rseed=None):
    """Train a neural network to detect variants.

    Parameters
    ----------
    train_file_path : str
        Input tabulated feature file for training.
    validate_file_path : str
        Input tabulated feature file for validation during training.
    model_file_path : str
        Output trained model.
    rseed : int
        Random seed to ensure reproducible results.  Set to zero for non-deterministic results.
    """
    if rseed:
        logging.info("************************************************************************************************")
        logging.info("NOTICE: setting the random seed also forces single-threaded execution to ensure reproducibility.")
        logging.info("************************************************************************************************")
        logging.debug("Setting random seed = %d" % rseed)
        random.seed(rseed)
        np.random.seed(rseed)
        tf.set_random_seed(rseed)

        # Limit operation to 1 thread for deterministic results.
        session_conf = tf.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)
        sess = tf.Session(graph=tf.get_default_graph(), config=session_conf)
        keras.backend.set_session(sess)
    else:
        logging.info("************************************************************************************************")
        logging.info("NOTICE: results are not reproducible when rseed is not set.")
        logging.info("************************************************************************************************")

    logging.debug("Kyos train, version %s" % __version__)
    logging.debug("Loading data...")
    data, one_hot_labels = load_train_data(train_file_path, first_ftr_idx, last_ftr_idx)
    data_validation, one_hot_label_validation = load_train_data(validate_file_path, first_ftr_idx, last_ftr_idx)

    logging.debug("Defining model...")
    model = Sequential()
    model.add(Dense(40, input_dim=num_input_features))
    model.add(Activation("relu"))
    model.add(Dropout(0.2))

    model.add(Dense(30))
    model.add(Activation("relu"))
    model.add(Dropout(0.2))

    model.add(Dense(30))
    model.add(Activation("relu"))
    model.add(Dropout(0.2))

    model.add(Dense(30))
    model.add(Activation("relu"))

    model.add(Dense(9, activation='softmax'))

    optimizer = keras.optimizers.RMSprop(lr=0.0005)

    logging.debug("Compiling model...")
    model.compile(optimizer=optimizer,
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])

    early_stopping_monitor = EarlyStopping(patience=10, restore_best_weights=True)
    callbacks = [early_stopping_monitor]

    logging.debug("Fitting model...")
    model.fit(data, one_hot_labels, validation_data=(data_validation, one_hot_label_validation), batch_size=100000, callbacks=callbacks, epochs=100, verbose=2)

    logging.debug("Saving model...")
    model.save(model_file_path)
    logging.debug("Training finished.")


def test(model_file_path, test_file_path, vcf_file_path=None):
    """Test a neural network variant detector.

    Parameters
    ----------
    model_file_path : str
        Input trained model.
    test_file_path : str
        Input tabulated feature file for testing.
    vcf_file_path : str, optional
        Optional output VCF file.
    """

    data, labels, reference = load_test_data(test_file_path, first_ftr_idx, last_ftr_idx)

    model = keras.models.load_model(model_file_path)

    prediction = model.predict(data)

    output = np.argmax(prediction, axis=1)

    df = pd.DataFrame({"output": output, "labels": labels, "reference": reference})

    df["Mutation"] = df.labels != df.reference

    df["Correct_call"] = df.output == df.labels
    df["TN"] = (~df.Mutation) & df.Correct_call
    df["TP"] = df.Mutation & df.Correct_call
    df["FP"] = (~df.Mutation) & (~df.Correct_call)
    df["FN"] = df.Mutation & (~df.Correct_call)

    false_positives = df["FP"].sum()
    false_negatives = df["FN"].sum()
    true_positives = df["TP"].sum()
    true_negatives = df["TN"].sum()

    print("FP:", false_positives, "FN:", false_negatives, "TP:", true_positives, "TN:", true_negatives)


def call(model_file_path, ftr_file_path, vcf_file_path):
    """Call variants from a tabulated feature file when the truth is unknown.

    Parameters
    ----------
    model_file_path : str
        Input trained model.
    ftr_file_path : str
        Input tabulated feature file.
    vcf_file_path : str, optional
        Optional output VCF file.
    """
    print('The "call" command is not implemented yet')



# -*- coding: utf-8 -*-

"""
test_cli
----------------------------------

Tests for cli module.
"""

import pytest

from kyos import cli


def test_error_on_empty_command_line():
    """Verify exception on empty command line."""
    with pytest.raises(SystemExit):
        cli.run_from_line("")



# -*- coding: utf-8 -*-

"""
test_nn
----------------------------------

Tests for `nn` module.
"""

from __future__ import absolute_import

import numpy as np

from kyos import features
from kyos import nn

import random

random.seed(1)


def test_load_train_data_all_classes(tmpdir):
    """Verify training features and targets are loaded properly when all
    output classes are represented in the data file.
    """

    # Header row. Names don't matter for this test
    columns = features.feature_names + [features.target_label_name]
    content = '\t'.join(columns) + '\n'

    # Feature data rows, one row for each possible target class
    for target in nn.output_classes:
        row = [str(i) for i, col in enumerate(features.feature_names)] + [target]
        content += '\t'.join(row)
        content += '\n'

    path = tmpdir.join("ftrFile.tsv")
    path.write(content)

    ftrs, targets = nn.load_train_data(str(path), 3, 5, scaling=None)

    assert(ftrs.dtype == np.float32)
    assert(targets.dtype == np.float32)

    assert(np.array_equal(ftrs, np.array(
      [[3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.]])))

    assert(np.array_equal(targets, np.array(
      [[1., 0., 0., 0., 0., 0., 0., 0., 0.],
       [0., 1., 0., 0., 0., 0., 0., 0., 0.],
       [0., 0., 1., 0., 0., 0., 0., 0., 0.],
       [0., 0., 0., 1., 0., 0., 0., 0., 0.],
       [0., 0., 0., 0., 1., 0., 0., 0., 0.],
       [0., 0., 0., 0., 0., 1., 0., 0., 0.],
       [0., 0., 0., 0., 0., 0., 1., 0., 0.],
       [0., 0., 0., 0., 0., 0., 0., 1., 0.],
       [0., 0., 0., 0., 0., 0., 0., 0., 1.]])))


def test_load_train_data_some_classes(tmpdir):
    """Verify training features and targets are loaded properly when some
    output classes are not represented in the data file.
    """

    # Header row. Names don't matter for this test
    columns = features.feature_names + [features.target_label_name]
    content = '\t'.join(columns) + '\n'

    # Feature data rows, one row for each possible target class
    row = [str(i) for i, col in enumerate(features.feature_names)] + [nn.output_classes[0]]
    content += '\t'.join(row)
    content += '\n'
    row = [str(i) for i, col in enumerate(features.feature_names)] + [nn.output_classes[-2]]
    content += '\t'.join(row)
    content += '\n'

    path = tmpdir.join("ftrFile.tsv")
    path.write(content)

    ftrs, targets = nn.load_train_data(str(path), 3, 5, scaling=None)

    assert(ftrs.dtype == np.float32)
    assert(targets.dtype == np.float32)

    assert(np.array_equal(ftrs, np.array(
      [[3., 4., 5.],
       [3., 4., 5.]])))

    assert(np.array_equal(targets, np.array(
      [[1., 0., 0., 0., 0., 0., 0., 0., 0.],
       [0., 0., 0., 0., 0., 0., 0., 1., 0.]])))


def test_load_test_data_all_classes(tmpdir):
    """Verify training features and targets are loaded properly when all
    output classes are represented in the data file.
    """

    # Header row. Names don't matter for this test
    columns = features.feature_names + [features.target_label_name]
    content = '\t'.join(columns) + '\n'

    # Feature data rows, one row for each possible target class
    for target in nn.output_classes:
        row = [str(i) for i, col in enumerate(features.feature_names[0:-1])] + random.sample(["A", "C", "T", "G"], 1) + [target]
        content += '\t'.join(row)
        content += '\n'

    path = tmpdir.join("ftrFile.tsv")
    path.write(content)

    ftrs, targets, refs = nn.load_test_data(str(path), 3, 5, scaling=None)

    assert(ftrs.dtype == np.float32)
    assert(targets.dtype == np.int64)
    assert(refs.dtype == np.int64)

    assert(np.array_equal(ftrs, np.array(
      [[3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.],
       [3., 4., 5.]])))

    assert(np.array_equal(targets, np.array(
      [0, 1, 2, 3, 4, 5, 6, 7, 8])))

    assert(np.array_equal(refs, np.array(
      [0, 3, 3, 2, 2, 2, 1, 3, 0])))



# -*- coding: utf-8 -*-

"""
This module extracts features useful for detecting variants from a BAM file.
"""

from __future__ import print_function
from __future__ import absolute_import

from Bio import SeqIO
from collections import Counter
from collections import namedtuple
import csv
import errno
import logging
import os
import pysam
import random
import sys

# Tuple to hold metrics per base
PerBaseTuple = namedtuple("PerBaseTuple", ['A', 'T', 'C', 'G', 'N', 'DEL', 'INS'])

# Tuple to hold truth dict value
TruthTuple = namedtuple("TruthTuple", ['ref_base', 'variant'])


feature_names = ["SampleName", "Chrom", "Position", "A", "T", "C", "G", "N", "a", "t", "c", "g", "n", "Insertion", "Deletion", "MapqA", "MapqT", "MapqC", "MapqG", "MapqN", "MapqDel", "MapqIns",
                 "BaseQualA", "BaseQualT", "BaseQualC", "BaseQualG", "BaseQualN", "BaseQualDel", "BaseQualIns", "RefBase"]

# Expose feature names to the nn module
first_ftr_name = "A"
last_ftr_name = "BaseQualIns"
read_count_feature_names = ["A", "T", "C", "G", "N", "a", "t", "c", "g", "n", "Insertion", "Deletion"]
map_quality_feature_names = ["MapqA", "MapqT", "MapqC", "MapqG", "MapqN", "MapqDel", "MapqIns"]
base_quality_feature_names = ["BaseQualA", "BaseQualT", "BaseQualC", "BaseQualG", "BaseQualN", "BaseQualDel", "BaseQualIns"]
target_label_name = "Truth"


def mkdir_p(path):
    """Python equivalent of bash mkdir -p.

    Parameters
    ----------
    path : str
        Directory path to create.

    Raises
    ------
    OSError if the directory does not already exist and cannot be created
    """
    if path == "":
        return
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def read_summary_file(summary_file):
    """Read a SNP Mutator summary file to get the truth alleles.

    The SNP Mutator summary file looks like this:

    Replicate              Position  OriginalBase  NewBase
    CFSAN000189_mutated_1  25183     g             T
    CFSAN000189_mutated_1  34442     a             aA_insertion
    CFSAN000189_mutated_1  42998     g             A
    CFSAN000189_mutated_1  92950     a             _deletion
    CFSAN000189_mutated_1  94829     g             C

    Parameters
    ----------
    summary_file : str
        Path to SNP Mutator summary file

    Returns
    ------
    dict
        key = str(sample_name+position)
        value = truth allele
    """
    truth = {}

    with open(summary_file, "r") as f:
        f.next()  # skip first row with headers
        for line in f:
            replicate, pos, ref_base, truth_variant = line.split()
            key = replicate + " " + pos
            if truth_variant.endswith("_insertion"):  # Strip off reference base from insertions
                truth_variant = truth_variant[1:]
            truth[key] = TruthTuple(ref_base.upper(), truth_variant)

    return truth


def any_variant_evidence(pileup_base_counter, ref_seq_dict, chrom, position):
    """Determine whether there is any evidence of variation occuring at the position determined by the pileup column.

    Parameters
    ----------
    pileup_base_counter : dict
        Contains the bases as the keys and the number of occurences as the values
    ref_seq_dict : dict of str
        Dictionary of contigs.  Key=contig name, value=sequence string.
    chrom : str
        Name of the contig
    position : int
        Zero-based position relative to the start of the contig

    Returns
    -------
    bool
        True if there is any evidence of variant, false if there is not evidence
    """
    if len(pileup_base_counter) == 0:
        return False

    ref_base = ref_seq_dict[chrom][position]

    keys = pileup_base_counter.keys()
    if len(keys) > 2:
        return True

    elif "+" in keys or "-" in keys or "*" in keys or "N" in keys or "n" in keys:
        return True
    elif len(keys) == 2:
        if keys[0].upper() != keys[1].upper():
            return True
        else:
            if keys[0].upper() != ref_base:
                return True
    else:
        if keys[0].upper() != ref_base:
            return True
    return False


def convert_to_csv(observations, file_path):
    """Writes the tabular data to a csv file

    Parameters
    ----------
    observations : list of dictionaries
        Contains the data of each mutated place as a row in the list
    file_path : string
        Path to the output file

    """
    out_dir = os.path.dirname(file_path)
    mkdir_p(out_dir)
    with open(file_path, "w") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=feature_names, delimiter='\t')

        writer.writeheader()
        for row in observations:
            writer.writerow(row)


def mean(qualities):
    """Calculates the average of a list which might be empty.

    Parameters
    ----------
    qualtiies : list
        list of numbers

    Returns
    -------
    float
        Float representing the mean numbers in the the list; zero if the list is empty.
    """
    if len(qualities) == 0:
        return 0

    de_phred = []

    for x in qualities:
        de_phred.append(100-(100*(10**(-1.0*x/10.0))))

    return sum(de_phred) / float(len(de_phred))


def get_average_qualities(pileup_alleles, pileup_qualities):
    """Gets the average quality for each A,T,C,G,N, insertions and deletions.

    Parameters
    ----------
    pileup_alleles : pileup column
        Pileup column containing a list of bases.
    pileup_qualities : pileup column
        Pileup column containing a list of quality scores for each allele.  The quality score
        could be either base quality or mapping quality.

    Returns
    -------
    PerBaseTuple
        Named tuple, PerBaseTuple, containing the average quality for each of (A,T,C,G,N,insertion,deletion)
    """
    qualA = []
    qualT = []
    qualC = []
    qualG = []
    qualN = []
    qualDel = []
    qualIns = []

    for x in range(len(pileup_alleles)):
        if pileup_alleles[x].upper() == "A":
            qualA.append(pileup_qualities[x])
        elif pileup_alleles[x].upper() == "T":
            qualT.append(pileup_qualities[x])
        elif pileup_alleles[x].upper() == "C":
            qualC.append(pileup_qualities[x])
        elif pileup_alleles[x].upper() == "G":
            qualG.append(pileup_qualities[x])
        elif pileup_alleles[x].upper() == "N":
            qualN.append(pileup_qualities[x])
        elif "+" in pileup_alleles[x]:
            qualIns.append(pileup_qualities[x])
        elif "-" in pileup_alleles[x] or "*" in pileup_alleles[x]:
            qualDel.append(pileup_qualities[x])
        else:
            print("Unknown pileup_allele", pileup_alleles[x], file=sys.stderr)

    return PerBaseTuple(mean(qualA), mean(qualT), mean(qualC), mean(qualG), mean(qualN), mean(qualDel), mean(qualIns))


def read_ref_file(reference_path):
    """Read a fasta reference and return a dictionary of contigs.

    Parameters
    ----------
    reference_path : str
        Path to reference fasta file.

    Returns
    -------
    ref_seq_dict : dict
        Dictionary with key=contig name and value=string of bases in the contig.
    """
    ref_seq_dict = {}
    for seqrecord in SeqIO.parse(reference_path, "fasta"):
        ref_seq_dict[seqrecord.id] = str(seqrecord.seq).upper()
    return ref_seq_dict


def create_tabular_data(input_file, output_file, reference_file, truth_file=None, tn_fraction=1.0, force_truth=None, rseed=None):
    # TODO: This function needs to also examine the summary file or the VCF file.
    #       This is the truth dataset.
    #       Every mutated position needs to be included in the output file, even if
    #       there is no evidence of variation at the position.
    # TODO: Determine if the pileup iterator skips some positions.  Could it skip
    #       a position in the truth dataset?
    """Read a BAM file, extract informative features, and write the features to
    a TSV file.

    Parameters
    ----------
    input_file : str
        Path to input BAM file
    output_file : str
        Path to output TSV file
    reference_file : str
        Path to reference fasta file
    truth_file : str
        Path to the file containing the truth allele
    tn_fraction : float
        Fraction of True Negatives to write to the output file.
    force_truth : bool
        Extract all known variant observations even if there is no coverage. Requires a summary file.
    rseed : int
        Random seed to ensure reproducible results.  Set to zero for non-deterministic results.
    """
    if rseed:
        random.seed(rseed)

    input_dir = os.path.dirname(input_file)
    sample_name = os.path.basename(input_dir)

    if truth_file:
        truth_dict = read_summary_file(truth_file)
        feature_names.append("Truth")

    ref_seq_dict = read_ref_file(reference_file)

    samfile = pysam.AlignmentFile(input_file, "rb")

    observations = []

    all_positions = set()

    for pileupcolumn in samfile.pileup(until_eof=True):

        min_base_quality = 13
        pileupcolumn.set_min_base_quality(min_base_quality)

        pileup_bases = pileupcolumn.get_query_sequences(mark_matches=True, add_indels=True)

        observation = {}  # dictionary of features keyed by feature_names

        for x in feature_names:
            if x not in observation:
                observation[x] = 0

        contig = pileupcolumn.reference_name

        observation["Chrom"] = contig
        observation["RefBase"] = ref_seq_dict[contig][pileupcolumn.pos]
        observation["SampleName"] = sample_name
        observation["Position"] = pileupcolumn.pos + 1
        observation["Truth"] = observation["RefBase"]

        if truth_file:
            key = sample_name + " " + str(pileupcolumn.pos + 1)
            mutated_position_flag = key in truth_dict
            if mutated_position_flag:
                # TODO: test this for insertions and deletions, not sure this will work properly
                observation["Truth"] = truth_dict[key].variant.upper()
            else:
                observation["Truth"] = observation["RefBase"]

        if len(pileup_bases) == 0:
            # no coverage at this position
            logging.debug("%d No coverage at base quality %d" % (pileupcolumn.pos + 1, min_base_quality))
            if force_truth and mutated_position_flag:
                observations.append(observation)
                all_positions.add(sample_name + " " + str(pileupcolumn.pos + 1))
                logging.debug("Added no-coverage known true mutation for %s at %d" % (sample_name, pileupcolumn.pos + 1))
            continue

        base_counts = Counter(pileup_bases)

        is_any_variant_evidence = any_variant_evidence(base_counts, ref_seq_dict, contig, pileupcolumn.pos)

        in_summary = force_truth and mutated_position_flag

        '''if in_summary:
            print(in_summary, pileupcolumn.pos, key)'''

        if is_any_variant_evidence or in_summary:
            if tn_fraction < 1.0 and not mutated_position_flag and random.random() > tn_fraction:
                continue

            for key in base_counts:
                if key in feature_names:
                    observation[key] = base_counts[key]
                else:
                    # A regex pattern `\+[0-9]+[ACGTNacgtn]+' indicates there is an insertion between this reference position
                    # and the next reference position. The length of the insertion is given by the integer in the pattern,
                    # followed by the inserted sequence.
                    if "+" in key:
                        observation["Insertion"] += 1

                    # A regex pattern `-[0-9]+[ACGTNacgtn]+' represents a deletion from the reference.
                    # An * (asterisk) is a placeholder for a deleted base in a multiple basepair deletion that was mentioned
                    # in a previous line by the -[0-9]+[ACGTNacgtn]+ notation.
                    elif "-" in key or "*" in key:
                        observation["Deletion"] += 1

                    else:
                        print(1 + pileupcolumn.pos, "Unexpected base:", key, pileup_bases, file=sys.stderr)

            mapping_qualities = pileupcolumn.get_mapping_qualities()
            base_qualities = pileupcolumn.get_query_qualities()

            ave_mapping_quality = get_average_qualities(pileup_bases, mapping_qualities)
            ave_base_quality = get_average_qualities(pileup_bases, base_qualities)

            observation["MapqA"] = ave_mapping_quality.A
            observation["MapqT"] = ave_mapping_quality.T
            observation["MapqC"] = ave_mapping_quality.C
            observation["MapqG"] = ave_mapping_quality.G
            observation["MapqN"] = ave_mapping_quality.N
            observation["MapqDel"] = ave_mapping_quality.DEL
            observation["MapqIns"] = ave_mapping_quality.INS

            observation["BaseQualA"] = ave_base_quality.A
            observation["BaseQualT"] = ave_base_quality.T
            observation["BaseQualC"] = ave_base_quality.C
            observation["BaseQualG"] = ave_base_quality.G
            observation["BaseQualN"] = ave_base_quality.N
            observation["BaseQualDel"] = ave_base_quality.DEL
            observation["BaseQualIns"] = ave_base_quality.INS

            # Fill in missing data
            for x in feature_names:
                if x not in observation:
                    observation[x] = 0

            observations.append(observation)
            all_positions.add(sample_name + " " + str(pileupcolumn.pos + 1))

    for key, value in truth_dict.iteritems():
        try:
            replicate, position_str = key.split()
            position = int(position_str)
            if sample_name == replicate and key not in all_positions:
                observation = {}  # dictionary of features keyed by feature_names

                for x in feature_names:
                    if x not in observation:
                        observation[x] = 0

                contig = ""

                observation["Chrom"] = "chrom"
                observation["RefBase"] = value.ref_base
                observation["SampleName"] = sample_name
                observation["Position"] = position
                observation["Truth"] = value.variant

                observations.append(observation)
                logging.debug("Added missing known true mutation for %s at %d" % (sample_name, position))
        except Exception:
            logging.exception("Exception with truth dict.  key: %s value: %s", repr(key), repr(value))

    samfile.close()

    logging.debug("Converting to csv...")
    convert_to_csv(observations, output_file)
    logging.debug("Tabulate finished.")


def merge(input_paths):
    """Merge multiple tabulated files and write to stdout.

    Parameters
    ----------
    input_paths : list of str
        List of tsv file paths
    """
    file_num = 0
    for path in input_paths:
        file_num += 1
        with open(path) as f:
            if file_num > 1:
                next(f)  # skip header line in all files after first
            for line in f:
                sys.stdout.write(line)

    sys.stdout.flush()
